// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
//
// BT709ToBT601 - Conv HDTV to SDTV color
//
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Version      Developer             Changes
//
// 07 Apr 2003   0.1.0.0      Tom Barry		        New 
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <crtdbg.h>

#ifdef AVISYNTH_20
#include "avisynth20.h"
#else
#include "avisynth.h"
#endif

#include <malloc.h>
#include <string.h>


const __int64 ZEROS = 0x0000000000000000;

class BT709ToBT601 : public GenericVideoFilter {
    int  Max_Frames;

    PVideoFrame src;
    PVideoFrame dst;

	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;

#ifndef AVISYNTH_20
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);
	PVideoFrame GetFrame_YV12(int inFrame, IScriptEnvironment* env);
#endif

public:
    BT709ToBT601(PClip _child, 
		IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_asm emms;

        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;
		SSE2enabled = (env->GetCPUFlags() & CPUF_SSE2) !=0;

		if ((vi.width % 8) || (vi.height % 2))
		{
			env->ThrowError("Sorry, BT709ToBT601 needs size multiples of 8x2");
		}

	    if (!vi.IsYV12())
			env->ThrowError("BT709ToBT601: Supports YV12 color format only");
	}	    
    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
};

// The following macro sets a value suitable for the pmulhw instruction, no range checks!
#define FAC(val) ( (unsigned short) (65536*val))

__declspec(align(16))
	const unsigned short fact_YU[4] = {FAC(.099312), FAC(.099312), FAC(.099312), FAC(.099312)};
__declspec(align(16))
	const unsigned short fact_YV[4] = {FAC(.191699), FAC(.191699), FAC(.191699), FAC(.191699)};

// fact_UU scaled down by half to avoid overflow
__declspec(align(16))
	const unsigned short fact_UU[4] = {FAC(.989854/2), FAC(.989854/2), FAC(.989854/2), FAC(.989854/2)}; 
__declspec(align(16))
	const unsigned short fact_UV[4] = {FAC(.110652), FAC(.110652), FAC(.110652), FAC(.110652)}; 	
__declspec(align(16))
	const unsigned short fact_VU[4] = {FAC(.072452), FAC(.072452), FAC(.072452), FAC(.072452)}; 
// fact_VV scaled down by half to avoid overflow
__declspec(align(16))
	const unsigned short fact_VV[4] = {FAC(.983398/2), FAC(.983398/2), FAC(.983398/2), FAC(.983398/2)};

const __int64 Q128 = 0x0080008000800080;

PVideoFrame __stdcall BT709ToBT601::GetFrame(int inFrame, IScriptEnvironment* env)
{
const BYTE* srcp_Y;			// Y
unsigned char* dstp_Y;
const BYTE* srcp_U;			// Cb
unsigned char* dstp_U;
const BYTE* srcp_V;			// Cr
unsigned char* dstp_V;
int src_pitch_Y;
int dst_pitch_Y;
int rowsize_Y;
int height_Y;

int src_pitch_UV;
int dst_pitch_UV;

int y;
int loopctr;
    src = child->GetFrame(inFrame, env);

	if (inFrame < 0 || inFrame >= Max_Frames)
	{
		return src;						// don't filter first and last frames
	}

    dst = env->NewVideoFrame(vi);

	srcp_Y = src->GetReadPtr(PLANAR_Y);
	dstp_Y = dst->GetWritePtr(PLANAR_Y);
	src_pitch_Y = src->GetPitch(PLANAR_Y);
	dst_pitch_Y = dst->GetPitch(PLANAR_Y);
	rowsize_Y = dst->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
	height_Y = dst->GetHeight(PLANAR_Y);

	// Next, the U plane
	srcp_U = src->GetReadPtr(PLANAR_U);
	dstp_U = dst->GetWritePtr(PLANAR_U);
	src_pitch_UV = src->GetPitch(PLANAR_U);
	dst_pitch_UV = dst->GetPitch(PLANAR_U);

	// And the V plane, same sizes as U, different locations
	srcp_V = src->GetReadPtr(PLANAR_V);   
	dstp_V = dst->GetWritePtr(PLANAR_V);   

	for (y=1; y <= height_Y/2; y++)			// doing 2 luma lines at a time
	{
		loopctr = rowsize_Y >> 3;
		
		__asm
		{
		mov		esi, srcp_Y
		mov		edi, dstp_Y
		mov		eax, srcp_U 
		mov		ebx, dstp_U 
		mov		ecx, srcp_V 
		mov		edx, dstp_V	
		pxor	mm7, mm7					// useful const zero
		movq	mm6, Q128					// " " 128
		
		align	16
ConvLoop:
		// get 4 bytes each of U, V
		movd	mm0, dword ptr[eax]			// move 1st 4 U bytes
		movd	mm1, dword ptr[ecx]			// move 1st 4 V bytes
		punpcklbw mm0, mm7					// make 4 words
		punpcklbw mm1, mm7					// make 4 words
		psubw   mm0, mm6					// adj for 128 chroma offset
		psubw   mm1, mm6					// adj for 128 chroma offset
		// First get Y adjustment
		movq	mm2, mm0
		movq	mm3, mm1
		pmulhw  mm2, fact_YU				// YU factor (U term in adjusted Y)
		pmulhw  mm3, fact_YV
		paddw   mm2, mm3					// total adjusted amount to add to Y
		movq	mm3, mm2
		punpcklwd mm2, mm2					// words <1,1,0,0>
		punpckhwd mm3, mm3					// words <3,3,2,2>

		// add to Y, top line of 2
		movq    mm4, qword ptr[esi]
		movq	mm5, mm4
		punpcklbw mm4, mm7
		punpckhbw mm5, mm7
		paddw	mm4, mm2
		paddw	mm5, mm3
		packuswb mm4, mm5					// pack back to 8 bytes, saturate to 0-255
		movq	qword ptr[edi], mm4

		// add to Y, bottom line of 2
		add		esi, src_pitch_Y			// bump to next line
		add		edi, dst_pitch_Y
		movq    mm4, qword ptr[esi]
		movq	mm5, mm4
		punpcklbw mm4, mm7
		punpckhbw mm5, mm7
		paddw	mm4, mm2
		paddw	mm5, mm3
		packuswb mm4, mm5					// pack back to 8 bytes, saturate to 0-255
		movq	qword ptr[edi], mm4
		sub		esi, src_pitch_Y			// restore curr line
		sub		edi, dst_pitch_Y

		// adjust U
		movq	mm2, mm0
		movq	mm3, mm1
		paddw	mm2, mm2					// adj for fact_UU being scaled down
		pmulhw  mm2, fact_UU				// UU factor (U term in adjusted U)
		pmulhw  mm3, fact_UV				// UV factor (V term in adjusted U)
		psubsw  mm2, mm3					// 2nd term negative, this is new U 
		paddw   mm2, mm6					// bias up by 128 again
		packuswb mm2, mm7					// back to 4 bytes 
		movd	dword ptr[ebx], mm2			// store adjusted U

		// adjust V
		movq	mm2, mm0
		movq	mm3, mm1
		pmulhw  mm2, fact_VU				// VU factor (U term in adjusted V)
		paddw	mm3, mm3					// adj for fact_VV being scaled down
		pmulhw  mm3, fact_VV				// VV factor (V term in adjusted V)
		psubsw  mm3, mm2					// 1st term negative, this is new V 
		paddw   mm3, mm6					// bias up by 128 again
		packuswb mm3, mm7					// back to 4 bytes 
		movd	dword ptr[edx], mm3			// store adjusted V

		add		esi, 8						// bump ptrs
		add		edi, 8						
		add		eax, 4						
		add		ebx, 4						
		add		ecx, 4						
		add		edx, 4						

		dec		loopctr
		jg		ConvLoop
		}

		// adjust for next line
		srcp_Y += 2*src_pitch_Y;
		dstp_Y += 2*dst_pitch_Y;
		srcp_U += src_pitch_UV;
		dstp_U += dst_pitch_UV;
		srcp_V += src_pitch_UV;
		dstp_V += dst_pitch_UV;
	}
	__asm emms;

    return dst;
}


AVSValue __cdecl Create_BT709ToBT601(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
	_asm emms;
    return new BT709ToBT601(args[0].AsClip(), env);
}

extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
	_asm emms;
    env->AddFunction("BT709ToBT601", "c", Create_BT709ToBT601, 0);
    return "`BT709ToBT601' Conv HDTV to SDTV color";
}

